import os
import shutil
from datetime import datetime

from dateutil.tz import tzlocal
from pynwb import NWBHDF5IO, NWBFile
from pynwb.ophys import (
    CorrectedImageStack,
    Fluorescence,
    ImageSegmentation,
    ImageSeries,
    MotionCorrection,
    OpticalChannel,
    RoiResponseSeries,
    TimeSeries,
    TwoPhotonSeries,
)

from optinist.api.nwb.nwb import NWBDATASET
from optinist.api.nwb.optinist_data import PostProcess


class NWBCreater:
    @classmethod
    def acquisition(cls, config):
        nwbfile = NWBFile(
            session_description=config["session_description"],
            identifier=config["identifier"],
            experiment_description=config["experiment_description"],
            session_start_time=datetime.now(tzlocal()),
        )

        # 顕微鏡情報を登録
        device = nwbfile.create_device(
            name=config["device"]["name"],
            description=config["device"]["description"],
            manufacturer=config["device"]["manufacturer"],
        )

        # 光チャネルを登録
        optical_channel = OpticalChannel(
            name=config["optical_channel"]["name"],
            description=config["optical_channel"]["description"],
            emission_lambda=float(config["optical_channel"]["emission_lambda"]),
        )

        # imaging planeを追加
        imaging_plane = nwbfile.create_imaging_plane(
            name=config["imaging_plane"]["name"],
            description=config["imaging_plane"]["description"],
            optical_channel=optical_channel,  # 光チャネル
            device=device,  # 電極デバイス
            imaging_rate=float(config["imaging_plane"]["imaging_rate"]),  # 画像の比率Hz
            excitation_lambda=float(
                config["imaging_plane"]["excitation_lambda"]
            ),  # 励起（れいき）波長
            indicator=config["imaging_plane"]["indicator"],  # カルシウムインディケーター
            location=config["imaging_plane"]["location"],
        )

        # using internal data. this data will be stored inside the NWB file
        if (
            NWBDATASET.IMAGE_SERIES in config
            and "external_file" in config[NWBDATASET.IMAGE_SERIES]
        ):
            # image_data = config[NWBDATASET.IMAGE_SERIES]['external_file'].data
            image_path = config[NWBDATASET.IMAGE_SERIES]["external_file"].path
            starting_frames = (
                config[NWBDATASET.IMAGE_SERIES]["starting_frame"]
                if "starting_frame" in config[NWBDATASET.IMAGE_SERIES]
                else None
            )

            if isinstance(image_path, list) and len(image_path) > 1:
                if starting_frames == 0 or starting_frames == [0]:
                    starting_frames = [0 for _ in range(len(image_path))]
                elif isinstance(starting_frames, str):
                    starting_frames = starting_frames.split(",")
                    starting_frames = list(map(int, starting_frames))

            image_series = TwoPhotonSeries(
                name="TwoPhotonSeries",
                starting_frame=starting_frames,
                external_file=image_path,
                imaging_plane=imaging_plane,
                starting_time=float(config[NWBDATASET.IMAGE_SERIES]["starting_time"]),
                rate=1.0,
                unit="normalized amplitude",
            )
            nwbfile.add_acquisition(image_series)

        nwbfile.create_processing_module(
            name="ophys", description="optical physiology processed data"
        )

        nwbfile.create_processing_module(name="optinist", description="description")
        cls.ophys(nwbfile)

        return nwbfile

    @classmethod
    def ophys(cls, nwbfile):
        img_seg = ImageSegmentation()
        nwbfile.processing["ophys"].add(img_seg)

        if "TwoPhotonSeries" in nwbfile.acquisition:
            reference_images = nwbfile.acquisition["TwoPhotonSeries"]

            img_seg.create_plane_segmentation(
                name="PlaneSegmentation",
                description="output",
                imaging_plane=nwbfile.imaging_planes["ImagingPlane"],
                reference_images=reference_images,
            )

        return nwbfile

    @classmethod
    def motion_correction(cls, nwbfile, mc_data, xy_trans_data):
        # image_data = mc_data.data
        image_path = mc_data.path
        corrected = ImageSeries(
            name="corrected",  # this must be named "corrected"
            # data=image_data,
            external_file=image_path,
            unit="na",
            format="external",
            starting_time=0.0,
            rate=1.0,
        )

        xy_translation = TimeSeries(
            name="xy_translation",
            data=xy_trans_data,
            unit="pixels",
            starting_time=0.0,
            rate=1.0,
        )

        corrected_image_stack = CorrectedImageStack(
            corrected=corrected,
            original=nwbfile.acquisition["TwoPhotonSeries"],
            xy_translation=xy_translation,
        )

        motion_correction = MotionCorrection(
            corrected_image_stacks=corrected_image_stack
        )
        nwbfile.processing["ophys"].add(motion_correction)

        return nwbfile

    @classmethod
    def column(cls, nwbfile, name, discription, data):
        data_interfaces = nwbfile.processing["ophys"].data_interfaces
        plane_seg = data_interfaces["ImageSegmentation"].plane_segmentations[
            "PlaneSegmentation"
        ]
        plane_seg.add_column(name, discription, data)

        return nwbfile

    @classmethod
    def roi(cls, nwbfile, roi_list):
        data_interfaces = nwbfile.processing["ophys"].data_interfaces
        plane_seg = data_interfaces["ImageSegmentation"].plane_segmentations[
            "PlaneSegmentation"
        ]

        for col in roi_list[0]:
            if col != "pixel_mask" and col not in plane_seg.colnames:
                plane_seg.add_column(col, f"{col} list")

        for col in roi_list:
            plane_seg.add_roi(**col)

        return nwbfile

    @classmethod
    def fluorescence(
        cls, nwbfile, table_name, region, name, data, unit, timestamps=None, rate=0.0
    ):
        data_interfaces = nwbfile.processing["ophys"].data_interfaces
        plane_seg = data_interfaces["ImageSegmentation"].plane_segmentations[
            "PlaneSegmentation"
        ]

        region_roi = plane_seg.create_roi_table_region(table_name, region=region)

        roi_resp_series = RoiResponseSeries(
            name=name,
            data=data,
            rois=region_roi,
            unit=unit,
            timestamps=timestamps,
            rate=float(rate),
        )

        fluo = Fluorescence(name=name, roi_response_series=roi_resp_series)

        nwbfile.processing["ophys"].add(fluo)

        return nwbfile

    @classmethod
    def timeseries(cls, nwbfile, key, value):
        timeseries_data = TimeSeries(
            name=key,
            data=value.data,
            unit="second",
            starting_time=0.0,
            rate=1.0,
        )

        nwbfile.processing["ophys"].add(timeseries_data)

        return nwbfile

    @classmethod
    def behavior(cls, nwbfile, key, value):
        timeseries_data = TimeSeries(
            name=key,
            data=value.data,
            unit="second",
            starting_time=0.0,
            rate=1.0,
        )

        nwbfile.processing["optinist"].add(timeseries_data)

        return nwbfile

    @classmethod
    def postprocess(cls, nwbfile, key, value):
        postprocess = PostProcess(
            name=key,
            data=value,
        )

        nwbfile.processing["optinist"].add_container(postprocess)

        return nwbfile

    @classmethod
    def reaqcuisition(cls, nwbfile):
        new_nwbfile = NWBFile(
            session_description=nwbfile.session_description,
            identifier=nwbfile.identifier,
            experiment_description=nwbfile.experiment_description,
            session_start_time=nwbfile.session_start_time,
        )

        devices = []
        for key in nwbfile.devices.keys():
            device = new_nwbfile.create_device(
                name=key,
                description=nwbfile.devices[key].description,
                manufacturer=nwbfile.devices[key].manufacturer,
            )
            devices.append(device)

        old_optical_channel = nwbfile.imaging_planes["ImagingPlane"].optical_channel[0]
        optical_channels = []
        for old_optical_channel in nwbfile.imaging_planes[
            "ImagingPlane"
        ].optical_channel:
            optical_channel = OpticalChannel(
                name=old_optical_channel.name,
                description=old_optical_channel.description,
                emission_lambda=old_optical_channel.emission_lambda,
            )
            optical_channels.append(optical_channel)

        imaging_planes = []
        for key in nwbfile.imaging_planes.keys():
            imaging_plane = new_nwbfile.create_imaging_plane(
                name=nwbfile.imaging_planes[key].name,
                description=nwbfile.imaging_planes[key].description,
                optical_channel=optical_channels,  # 光チャネル
                device=devices[0],  # 電極デバイス
                imaging_rate=float(
                    nwbfile.imaging_planes["ImagingPlane"].imaging_rate
                ),  # 画像の比率Hz
                excitation_lambda=float(
                    nwbfile.imaging_planes["ImagingPlane"].excitation_lambda
                ),  # 励起（れいき）波長
                indicator=nwbfile.imaging_planes[
                    "ImagingPlane"
                ].indicator,  # カルシウムインディケーター
                location=nwbfile.imaging_planes["ImagingPlane"].location,
            )
            imaging_planes.append(imaging_plane)

        image_series = TwoPhotonSeries(
            name="TwoPhotonSeries",
            starting_frame=nwbfile.acquisition["TwoPhotonSeries"].starting_frame,
            external_file=nwbfile.acquisition["TwoPhotonSeries"].external_file,
            imaging_plane=imaging_planes[0],
            starting_time=nwbfile.acquisition["TwoPhotonSeries"].starting_time,
            rate=nwbfile.acquisition["TwoPhotonSeries"].rate,
            unit=nwbfile.acquisition["TwoPhotonSeries"].unit,
        )

        new_nwbfile.add_acquisition(image_series)

        new_nwbfile.create_processing_module(
            name="ophys", description="optical physiology processed data"
        )
        new_nwbfile.create_processing_module(name="optinist", description="description")
        cls.ophys(new_nwbfile)
        return new_nwbfile


def save_nwb(save_path, input_config, config):
    nwbfile = NWBCreater.acquisition(input_config)

    if NWBDATASET.POSTPROCESS in config:
        for key, value in config[NWBDATASET.POSTPROCESS].items():
            NWBCreater.postprocess(nwbfile, key, value)

    if NWBDATASET.TIMESERIES in config:
        for key, value in config[NWBDATASET.TIMESERIES].items():
            NWBCreater.timeseries(nwbfile, key, value)

    if NWBDATASET.BEHAVIOR in config:
        for key, value in config[NWBDATASET.BEHAVIOR].items():
            NWBCreater.behavior(nwbfile, key, value)

    if NWBDATASET.MOTION_CORRECTION in config:
        for mc in config[NWBDATASET.MOTION_CORRECTION].values():
            nwbfile = NWBCreater.motion_correction(nwbfile, **mc)

    if NWBDATASET.ROI in config:
        for roi_list in config[NWBDATASET.ROI].values():
            NWBCreater.roi(nwbfile, roi_list)

    if NWBDATASET.COLUMN in config:
        for value in config[NWBDATASET.COLUMN].values():
            nwbfile = NWBCreater.column(nwbfile, **value)

    if NWBDATASET.FLUORESCENCE in config:
        for value in config[NWBDATASET.FLUORESCENCE].values():
            nwbfile = NWBCreater.fluorescence(nwbfile, **value)

    with NWBHDF5IO(save_path, "w") as f:
        f.write(nwbfile)


def overwrite_nwb(config, save_path, nwb_file_name):
    # バックアップファイルを作成
    nwb_path = os.path.join(save_path, nwb_file_name)
    tmp_nwb_path = os.path.join(save_path, "tmp_" + nwb_file_name)

    # NWBファイルの読み込み
    with NWBHDF5IO(nwb_path, "r") as io:
        nwbfile = io.read()
        # acquisition を元ファイルから作成する
        new_nwbfile = NWBCreater.reaqcuisition(nwbfile)

        if NWBDATASET.POSTPROCESS in config:
            for key, value in config[NWBDATASET.POSTPROCESS].items():
                new_nwbfile = NWBCreater.postprocess(new_nwbfile, key, value)

        if NWBDATASET.TIMESERIES in config:
            for key, value in config[NWBDATASET.TIMESERIES].items():
                new_nwbfile = NWBCreater.timeseries(new_nwbfile, key, value)

        if NWBDATASET.BEHAVIOR in config:
            for key, value in config[NWBDATASET.BEHAVIOR].items():
                new_nwbfile = NWBCreater.behavior(new_nwbfile, key, value)

        if NWBDATASET.MOTION_CORRECTION in config:
            for mc in config[NWBDATASET.MOTION_CORRECTION].values():
                new_nwbfile = NWBCreater.motion_correction(new_nwbfile, **mc)

        if NWBDATASET.ROI in config:
            for roi_list in config[NWBDATASET.ROI].values():
                new_nwbfile = NWBCreater.roi(new_nwbfile, roi_list)

        if NWBDATASET.COLUMN in config:
            for value in config[NWBDATASET.COLUMN].values():
                new_nwbfile = nwbfile = NWBCreater.column(new_nwbfile, **value)

        if NWBDATASET.FLUORESCENCE in config:
            for value in config[NWBDATASET.FLUORESCENCE].values():
                new_nwbfile = NWBCreater.fluorescence(new_nwbfile, **value)

        with NWBHDF5IO(tmp_nwb_path, "w") as io:
            io.write(new_nwbfile)
    shutil.copyfile(tmp_nwb_path, nwb_path)
    os.remove(tmp_nwb_path)


def merge_nwbfile(old_nwbfile, new_nwbfile):
    for pattern in [
        NWBDATASET.POSTPROCESS,
        NWBDATASET.TIMESERIES,
        NWBDATASET.MOTION_CORRECTION,
        NWBDATASET.ROI,
        NWBDATASET.COLUMN,
        NWBDATASET.FLUORESCENCE,
        NWBDATASET.BEHAVIOR,
        NWBDATASET.IMAGE_SERIES,
    ]:
        if pattern in old_nwbfile and pattern in new_nwbfile:
            old_nwbfile[pattern].update(new_nwbfile[pattern])
        elif pattern in new_nwbfile:
            old_nwbfile[pattern] = new_nwbfile[pattern]

    return old_nwbfile



import logging
import os
from typing import Dict

from optinist.api.dir_path import DIRPATH
from optinist.api.utils.filepath_creater import create_directory, join_filepath


def get_logger(unique_id: str) -> logging.Logger:
    output_dirpath = join_filepath(
        [
            DIRPATH.OUTPUT_DIR,
            unique_id,
        ]
    )
    create_directory(output_dirpath)

    filepath = f"{output_dirpath}/error.log"
    if os.path.exists(filepath):
        try:
            os.remove(filepath)
        except Exception as e:
            print("[Exception][Logger]", e)

    logger = logging.getLogger(unique_id)

    # FileHandlerの設定
    fh = logging.FileHandler(filepath)
    fh.setLevel(logging.ERROR)
    fmt = logging.Formatter("%(asctime)s : %(levelname)s - %(filename)s - %(message)s")
    fh.setFormatter(fmt)
    logger.addHandler(fh)

    return logger


class Logger:
    def __init__(self, unique_id):
        self.unique_id = unique_id
        self.logger: logging.Logger = get_logger(unique_id)

    def smk_logger(self, msg: Dict[str, str] = None):
        """
        msg:
            level:
                "job_info": jobの始まりを通知。inputやoutputがある
                "job_finished": jobの終了を通知。
        """
        # pass
        # # エラーした
        # if "exception" in msg:
        #     self.logger.error(msg)

        if "level" in msg and "debug" in msg["level"]:
            level = msg["level"]
            if "debug" in level and "msg" in msg:
                if "Traceback" in msg["msg"]:
                    self.logger.error(msg)



from typing import List

import h5py
from fastapi import APIRouter

from optinist.api.dir_path import DIRPATH
from optinist.api.utils.filepath_creater import join_filepath
from optinist.routers.model import HDF5Node

router = APIRouter()


class HDF5Getter:
    @classmethod
    def get(cls, filepath) -> List[HDF5Node]:
        cls.hdf5_list = []
        with h5py.File(filepath, "r") as f:
            f.visititems(cls.get_ds_dictionaries)

        return cls.hdf5_list

    @classmethod
    def get_ds_dictionaries(cls, path: str, node: h5py.Dataset):
        if isinstance(node, h5py.Dataset):
            if len(node.shape) != 0:
                cls.recursive_dir_tree(cls.hdf5_list, path.split("/"), node, "")

    @classmethod
    def recursive_dir_tree(
        cls,
        node_list: List[HDF5Node],
        path_list: List[str],
        node: h5py.Dataset,
        parent_path: str,
    ):
        name = path_list[0]
        path = name if parent_path == "" else f"{parent_path}/{name}"

        is_exists = False
        # 既にkeyがある
        for i, value in enumerate(node_list):
            if value.name == name:
                is_exists = True
                if len(path_list) > 1:
                    cls.recursive_dir_tree(
                        node_list[i].nodes, path_list[1:], node, path
                    )

        if not is_exists:
            if len(path_list) > 1:
                node_list.append(
                    HDF5Node(
                        isDir=True,
                        name=name,
                        path=path,
                        nodes=[],
                    )
                )
                cls.recursive_dir_tree(node_list[-1].nodes, path_list[1:], node, path)
            else:
                node_list.append(
                    HDF5Node(
                        isDir=False,
                        name=name,
                        path=path,
                        shape=node.shape,
                        nbytes=f"{int(node.nbytes / (1000**2))} M",
                    )
                )


@router.get("/hdf5/{file_path:path}", response_model=List[HDF5Node], tags=["outputs"])
async def get_files(file_path: str):
    file_path = join_filepath([DIRPATH.INPUT_DIR, file_path])
    return HDF5Getter.get(file_path)



from typing import Dict

import yaml

from optinist.api.experiment.experiment import ExptConfig, ExptFunction
from optinist.api.workflow.workflow import Edge, Node, NodeData, NodePosition, Style


class ExptConfigReader:
    @classmethod
    def read(cls, filepath) -> ExptConfig:
        with open(filepath, "r") as f:
            config = yaml.safe_load(f)

        return ExptConfig(
            unique_id=config["unique_id"],
            name=config["name"],
            timestamp=config["timestamp"],
            hasNWB=config["hasNWB"],
            function=cls.read_function(config["function"]),
            nodeDict=cls.read_nodeDict(config["nodeDict"]),
            edgeDict=cls.read_edgeDict(config["edgeDict"]),
        )

    @classmethod
    def read_function(cls, config) -> Dict[str, ExptFunction]:
        return {
            key: ExptFunction(
                unique_id=value["unique_id"],
                name=value["name"],
                success=value["success"],
                hasNWB=value["hasNWB"],
            )
            for key, value in config.items()
        }

    @classmethod
    def read_nodeDict(cls, config) -> Dict[str, Node]:
        return {
            key: Node(
                id=key,
                type=value["type"],
                data=NodeData(**value["data"]),
                position=NodePosition(**value["position"]),
                style=Style(**value["style"]),
            )
            for key, value in config.items()
        }

    @classmethod
    def read_edgeDict(cls, config) -> Dict[str, Edge]:
        return {
            key: Edge(
                id=key,
                type=value["type"],
                animated=value["animated"],
                source=value["source"],
                sourceHandle=value["sourceHandle"],
                target=value["target"],
                targetHandle=value["targetHandle"],
                style=Style(**value["style"]),
            )
            for key, value in config.items()
        }

    @classmethod
    def rename(cls, filepath, new_name: str) -> ExptConfig:
        with open(filepath, "r") as f:
            config = yaml.safe_load(f)
            config["name"] = new_name

        with open(filepath, "w") as f:
            yaml.dump(config, f)

        return ExptConfig(
            unique_id=config["unique_id"],
            name=config["name"],
            timestamp=config["timestamp"],
            hasNWB=config["hasNWB"],
            function=cls.read_function(config["function"]),
            nodeDict=cls.read_nodeDict(config["nodeDict"]),
            edgeDict=cls.read_edgeDict(config["edgeDict"]),
        )



import shutil
from glob import glob
from typing import Dict

from fastapi import APIRouter
from fastapi.responses import FileResponse

from optinist.api.dir_path import DIRPATH
from optinist.api.experiment.experiment import ExptConfig, ExptImportData
from optinist.api.experiment.experiment_reader import ExptConfigReader
from optinist.api.utils.filepath_creater import join_filepath
from optinist.routers.model import DeleteItem, RenameItem

router = APIRouter()


@router.get("/experiments", response_model=Dict[str, ExptConfig], tags=["experiments"])
async def get_experiments():
    exp_config = {}
    config_paths = glob(
        join_filepath([DIRPATH.OUTPUT_DIR, "*", DIRPATH.EXPERIMENT_YML])
    )
    for path in config_paths:
        try:
            config = ExptConfigReader.read(path)
            config.nodeDict = []
            config.edgeDict = []
            exp_config[config.unique_id] = config
        except Exception:
            pass

    return exp_config


@router.patch(
    "/experiments/{unique_id}/rename", response_model=ExptConfig, tags=["experiments"]
)
async def rename_experiment(unique_id: str, item: RenameItem):
    config = ExptConfigReader.rename(
        join_filepath([DIRPATH.OUTPUT_DIR, unique_id, DIRPATH.EXPERIMENT_YML]),
        new_name=item.new_name,
    )
    config.nodeDict = []
    config.edgeDict = []

    return config


@router.get(
    "/experiments/import/{unique_id}",
    response_model=ExptImportData,
    tags=["experiments"],
)
async def import_experiment(unique_id: str):
    config = ExptConfigReader.read(
        join_filepath([DIRPATH.OUTPUT_DIR, unique_id, DIRPATH.EXPERIMENT_YML])
    )
    return {
        "nodeDict": config.nodeDict,
        "edgeDict": config.edgeDict,
    }


@router.delete("/experiments/{unique_id}", response_model=bool, tags=["experiments"])
async def delete_experiment(unique_id: str):
    try:
        shutil.rmtree(join_filepath([DIRPATH.OUTPUT_DIR, unique_id]))
        return True
    except Exception:
        return False


@router.post("/experiments/delete", response_model=bool, tags=["experiments"])
async def delete_experiment_list(deleteItem: DeleteItem):
    try:
        [
            shutil.rmtree(join_filepath([DIRPATH.OUTPUT_DIR, uid]))
            for uid in deleteItem.uidList
        ]
        return True
    except Exception:
        return False


@router.get("/experiments/download/nwb/{unique_id}", tags=["experiments"])
async def download_nwb_experiment(unique_id: str):
    nwb_path_list = glob(join_filepath([DIRPATH.OUTPUT_DIR, unique_id, "*.nwb"]))
    if len(nwb_path_list) > 0:
        return FileResponse(nwb_path_list[0])
    else:
        return False


@router.get("/experiments/download/nwb/{unique_id}/{function_id}", tags=["experiments"])
async def download_nwb_experiment_with_function_id(unique_id: str, function_id: str):
    nwb_path_list = glob(
        join_filepath([DIRPATH.OUTPUT_DIR, unique_id, function_id, "*.nwb"])
    )
    if len(nwb_path_list) > 0:
        return FileResponse(nwb_path_list[0])
    else:
        return False


@router.get("/experiments/download/config/{unique_id}", tags=["experiments"])
async def download_config_experiment(unique_id: str):
    config_filepath = join_filepath(
        [DIRPATH.OUTPUT_DIR, unique_id, DIRPATH.SNAKEMAKE_CONFIG_YML]
    )
    return FileResponse(config_filepath)



from typing import Any, Dict

from fastapi import APIRouter

from optinist.api.config.config_reader import ConfigReader
from optinist.api.dir_path import DIRPATH
from optinist.api.utils.filepath_creater import join_filepath
from optinist.routers.model import NWBParams, SnakemakeParams

router = APIRouter()


@router.get("/params/{name}", response_model=Dict[str, Any], tags=["params"])
async def get_params(name: str):
    filepath = join_filepath([DIRPATH.CONFIG_DIR, f"{name}.yaml"])
    config = ConfigReader.read(filepath)
    return config


@router.get("/snakemake", response_model=SnakemakeParams, tags=["params"])
async def get_snakemake_params():
    filepath = join_filepath([DIRPATH.CONFIG_DIR, "snakemake.yaml"])
    return ConfigReader.read(filepath)


@router.get("/nwb", response_model=NWBParams, tags=["params"])
async def get_nwb_params():
    filepath = join_filepath([DIRPATH.CONFIG_DIR, "nwb.yaml"])
    return ConfigReader.read(filepath)



from optinist.api.dataclass.dataclass import NWBFile, TimeSeriesData


def cell_grouping(
    neural_data: TimeSeriesData,
    output_dir: str,
    nwbfile: NWBFile = None,
    params: dict = None,
) -> dict():
    import numpy as np

    neural_data = neural_data.data
    std = neural_data.std

    if params["transpose"]:
        neural_data = neural_data.transpose()

    baseline = np.mean(neural_data, axis=1, keepdims=True)
    std = neural_data[np.argmax(neural_data, axis=1)].std()

    grouped_cells = (neural_data.max(axis=1) - baseline) / std > params["threshold"]

    info = {}
    info["grouped_cells"] = TimeSeriesData(
        neural_data[grouped_cells], std=std, file_name="grouped_cells_mean"
    )

    return info



import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
exec(open(os.path.join(here, "optinist", "version.py")).read())


def _requires_from_file(filepath):
    def take_package_name(name):
        # if "snakemake" in name:
        #     return (
        #         "snakemake @ https://github.com/ShogoAkiyama/snakemake/"
        #         "archive/refs/tags/v7.7.2-post1.zip"
        #     )
        # else:
        return name.strip()

    with open(filepath) as fp:
        return [take_package_name(pkg_name) for pkg_name in fp.readlines()]


setup(
    name="optinist",
    version=VERSION,  # noqa: F821
    description="Calcium Imaging Pipeline Tool",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/oist/optinist",
    author="OIST",
    license="GPL3.0",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
    ],
    install_requires=_requires_from_file("requirements.txt"),
    packages=find_packages(exclude=["optinist/tests*"]),
    tests_require=["pytest"],
    zip_safe=False,
    include_package_data=True,
    package_data={
        "": [
            "frontend/build/*",
            "frontend/build/static/*",
            "frontend/build/static/css/*",
            "frontend/build/static/js/*",
            "frontend/build/static/media/*",
            "app_config/*.yaml",
            "config/*.yaml",
            "conda/*.yaml",
            "Snakefile",
        ]
    },
    py_modules=["Snakefile"],
    entry_points={
        "console_scripts": [
            "run_optinist=optinist.__main__:main",
        ]
    },
)



from optinist.wrappers.lccd_wrapper.lccd_detection import lccd_detect

lccd_wrapper_dict = {
    "lccd": {
        "lccd_cell_detection": {
            "function": lccd_detect,
            "conda_name": "lccd",
            "conda_yaml": "lccd_env.yaml",
        },
    }
}



import os

from optinist.api.dataclass.dataclass import FluoData, LccdData, RoiData
from optinist.api.edit_ROI.wrappers.lccd_edit_roi.utils import set_nwbfile


def excute_delete_roi(node_dirpath, ids):
    import numpy as np

    lccd_data = np.load(
        os.path.join(node_dirpath, "lccd.npy"), allow_pickle=True
    ).item()

    images = lccd_data.get("images", None)
    roi = lccd_data.get("roi", None)
    is_cell = lccd_data.get("is_cell")
    delete_roi = lccd_data.get("delete_roi", [])

    shape = images.shape[:2]
    num_frames = images.shape[2]
    dff_f0_frames = 100
    dff_f0_percentile = 8

    # delete ROI
    is_cell[ids] = False
    delete_roi += [(id + 1) for id in ids]

    num_cell = roi.shape[1]
    images = images.reshape([images.shape[0] * images.shape[1], images.shape[2]])
    timeseries = np.zeros([num_cell, num_frames])

    # Get ROI list and extract Fluor
    im = []
    for i in range(num_cell):
        im.append((roi[:, i].reshape(shape)) * (i + 1))
        timeseries[i, :] = np.mean(images[roi[:, i] > 0, :], axis=0)
    if len(im) == 0:
        im = np.zeros((1, *shape))
    else:
        im = np.stack(im)
    im[im == 0] = np.nan

    # Get DFF
    timeseries_dff = np.ones([num_cell, num_frames]) * np.nan
    for i in range(num_cell):
        for k in range(num_frames):
            if (k - dff_f0_frames >= 0) and (k + dff_f0_frames < num_frames):
                f0 = np.percentile(
                    timeseries[i, k - dff_f0_frames : k + dff_f0_frames],
                    dff_f0_percentile,
                )
                timeseries_dff[i, k] = (timeseries[i, k] - f0) / f0

    lccd_data["roi"] = roi
    lccd_data["is_cell"] = is_cell
    lccd_data["delete_roi"] = delete_roi

    roi_list = [{"image_mask": roi[:, i].reshape(shape)} for i in range(num_cell)]

    info = {
        "lccd": LccdData(lccd_data),
        "cell_roi": RoiData(
            np.nanmax(im[is_cell], axis=0),
            output_dir=node_dirpath,
            file_name="cell_roi",
        ),
        "fluorescence": FluoData(timeseries, file_name="fluorescence"),
        "dff": FluoData(timeseries_dff, file_name="dff"),
        "nwbfile": set_nwbfile(lccd_data, roi_list),
    }

    return info



from dataclasses import dataclass, field
from typing import Dict, List, Union

from pydantic import BaseModel


@dataclass
class Rule:
    input: list
    return_arg: Union[str, Dict[str, str]]
    params: dict
    output: str
    type: str
    nwbfile: dict = None
    hdf5Path: str = None
    path: str = None


@dataclass
class FlowConfig:
    rules: Dict[str, Rule]
    last_output: list


class ForceRun(BaseModel):
    nodeId: str
    name: str


@dataclass
class SmkParam:
    use_conda: bool
    cores: int
    forceall: bool
    forcetargets: bool
    lock: bool
    forcerun: List[ForceRun] = field(default_factory=list)



import numpy as np
import scipy.sparse

from optinist.wrappers.lccd_wrapper.lccd_python import utils


def filter_roi_by_area(roi, min_area, max_area):
    """
    Parameters
    ----------
    roi: np.array or scipy.sparse.csc (Maybe work with other sparse matrix types)
        roi has shape (X, n).
        X is spacial dimension.
        n is the number of rois
    min_area: int
    max_area: int

    Returns
    -------
    roi
    """

    n_rois = roi.shape[1]
    indices = []
    for i in range(n_rois):
        area = np.sum(roi[:, i])
        if min_area <= area and area <= max_area:
            indices.append(i)
    return roi[:, np.array(indices)]


class RoiIntegration:
    def __init__(self, overlap_threshold, min_area, max_area, sparse=True, **kwargs):
        self.overlap_threshold = overlap_threshold
        self.min_area = min_area
        self.max_area = max_area
        self.sparse = sparse

    def delete_col(self, arr, idx):
        if self.sparse:
            return utils.delete_col_csc(arr, idx)
        else:
            return np.delete(arr, idx, 1)

    def cat_rois(self, roi_a, roi_b):
        if self.sparse:
            return scipy.sparse.hstack((roi_a, roi_b))
        return np.hstack((roi_a, roi_b))

    def get_area(self, roi, i):
        """
        get area of ith region in roi.
        i.e. get sum of ith column of 2d array.
        """
        if self.sparse:
            return utils.count_nonzero_values_in_col_csc(roi, i)
        return np.sum(
            roi[:, i]
        )  # for dense array, np.count_nonzero is slower than np.sum(col)

    def gather_overlapping_regions(self, sim):
        """
        Gather regions those have overlapping area with other regions.
        """
        if self.sparse:
            return [
                region_in_a
                for region_in_a in range(sim.shape[0])
                if sim.getrow(region_in_a).nnz > 0
            ]
        return [
            region_in_a
            for region_in_a in range(sim.shape[0])
            if np.sum(sim[region_in_a]) > 0
        ]

    def remove_overlap(self, roi_a, roi_b, i, j):
        """
        This is inplace operation.
        From both roi_a's ith region and roi_b's jth region,
        remove their intersection.

        Let
        roi_a[:, i] = [1, 0, 1, 1, 0, 1, 1].T
        roi_b[:, j] = [0, 1, 1, 0, 1, 1, 0].T

        intersection is [0, 0, 1, 0, 0, 1, 0].T
        then, calling remove_overlap(roi_a, roi_b, i, j),
        roi_a[:, i] = [1, 0, 0, 1, 0, 0, 1].T
        roi_b[:, j] = [0, 1, 0, 0, 1, 0, 0].T
        """

        if self.sparse:  # TODO: improve readability
            roi_a_region_indices = roi_a.indices[roi_a.indptr[i] : roi_a.indptr[i + 1]]
            roi_b_region_indices = roi_b.indices[roi_b.indptr[j] : roi_b.indptr[j + 1]]
            overlap_indices = utils.intersect_unique_sorted_1d(
                roi_a_region_indices, roi_b_region_indices
            )
            roi_a_region_separated_indices = np.setdiff1d(
                roi_a_region_indices, overlap_indices, True
            )  # This set operation may be slow.
            roi_b_region_separated_indices = np.setdiff1d(
                roi_b_region_indices, overlap_indices, True
            )
            utils.delete_col_csc_inplace(roi_a, i)
            utils.insert_binary_col_binary_csc(roi_a, roi_a_region_separated_indices, i)
            utils.delete_col_csc_inplace(roi_b, j)
            utils.insert_binary_col_binary_csc(roi_b, roi_b_region_separated_indices, j)
        else:
            region_non_overlap = (roi_a[:, i] * roi_b[:, j]) == 0
            region_non_overlap = region_non_overlap.astype(roi_a.dtype)
            roi_a[:, i] *= region_non_overlap
            roi_b[:, j] *= region_non_overlap

    def apply(self, roi_a, roi_b):
        # requires test
        if roi_b is None:
            return roi_a
        dtype = np.float32
        roi_a = roi_a.astype(dtype)
        roi_b = roi_b.astype(dtype)
        if self.sparse:
            roi_a = scipy.sparse.csc_matrix(roi_a)
            roi_b = scipy.sparse.csc_matrix(roi_b)
        while 1:
            change_occured = False
            sim = roi_a.T.dot(roi_b)
            # sim[i][j] is area of overlap
            # by ith region in roi_a and jth region in roi_b
            # roi_a.shape = (1500x1500, 5000), roi_b.shape = (1500x1500, 3000)
            # density 30 / (1500x1500)
            # roi_a.T.dot(roi_b) took 6ms (AMD Ryzen 5900x)

            overlapping_regions_in_a = self.gather_overlapping_regions(sim)
            for region_in_roi_a in overlapping_regions_in_a:
                region_in_roi_b = np.argmax(
                    sim[region_in_roi_a]
                )  # most overlapping region in roi_b
                area_overlap = sim[region_in_roi_a, region_in_roi_b]
                if area_overlap == 0:
                    continue
                area_a = self.get_area(roi_a, region_in_roi_a)
                area_b = self.get_area(roi_b, region_in_roi_b)
                if (
                    area_overlap / area_a > self.overlap_threshold
                    or area_overlap / area_b > self.overlap_threshold
                ):
                    change_occured = True
                    if area_a > area_b:
                        roi_b = self.delete_col(roi_b, region_in_roi_b)
                        # column deletion from large array is slow.
                    else:
                        roi_a = self.delete_col(roi_a, region_in_roi_a)
                    break
                elif (
                    area_overlap / area_a <= self.overlap_threshold
                    and area_overlap / area_b <= self.overlap_threshold
                ):
                    change_occured = True
                    self.remove_overlap(roi_a, roi_b, region_in_roi_a, region_in_roi_b)
                    break
                else:
                    continue
            if not change_occured:
                break
        roi = self.cat_rois(roi_a, roi_b)
        roi = filter_roi_by_area(roi, self.min_area, self.max_area)
        return roi



from optinist.api.dataclass.dataclass import FluoData, Suite2pData
from optinist.api.nwb.nwb import NWBDATASET


def suite2p_spike_deconv(
    ops: Suite2pData, output_dir: str, params: dict = None
) -> dict(ops=Suite2pData, spks=FluoData):
    import numpy as np
    from suite2p import default_ops, extraction

    print("start suite2_spike_deconv")
    ops = ops.data

    ops = {**default_ops(), **ops, **params}

    dF = ops["F"].copy() - ops["neucoeff"] * ops["Fneu"]
    dF = extraction.preprocess(
        F=dF,
        baseline=ops["baseline"],
        win_baseline=ops["win_baseline"],
        sig_baseline=ops["sig_baseline"],
        fs=ops["fs"],
        prctile_baseline=ops["prctile_baseline"],
    )
    spks = extraction.oasis(
        F=dF, batch_size=ops["batch_size"], tau=ops["tau"], fs=ops["fs"]
    )

    ops["spks"] = spks

    # NWBを追加
    nwbfile = {}

    # roiを追加
    stat = ops["stat"]
    roi_list = []
    for i in range(len(stat)):
        kargs = {}
        kargs["pixel_mask"] = np.array(
            [stat[i]["ypix"], stat[i]["xpix"], stat[i]["lam"]]
        ).T
        roi_list.append(kargs)

    nwbfile[NWBDATASET.ROI] = {"roi_list": roi_list}

    # Fluorenceを追加
    name = "Deconvolved"
    nwbfile[NWBDATASET.FLUORESCENCE] = {
        name: {
            "table_name": name,
            "region": list(range(len(spks))),
            "name": name,
            "data": spks,
            "unit": "lumens",
            "rate": ops["fs"],
        }
    }

    info = {
        "ops": Suite2pData(ops),
        "spks": FluoData(spks, file_name="spks"),
        "nwbfile": nwbfile,
    }

    return info



from optinist.api.edit_ROI.edit_ROI import ACTION, EditROI, EditRoiUtils
from optinist.api.edit_ROI.wrappers import edit_roi_wrapper_dict

__all__ = ["ACTION", "EditROI", "EditRoiUtils", "edit_roi_wrapper_dict"]



from dataclasses import dataclass
from typing import Dict, Optional

from optinist.api.workflow.workflow import Edge, Node


@dataclass
class ExptFunction:
    unique_id: str
    name: str
    success: str
    hasNWB: bool


@dataclass
class ExptConfig:
    timestamp: str
    name: Optional[str]
    unique_id: str
    hasNWB: bool
    function: Dict[str, ExptFunction]
    nodeDict: Dict[str, Node]
    edgeDict: Dict[str, Edge]


@dataclass
class ExptImportData:
    nodeDict: Dict[str, Node]
    edgeDict: Dict[str, Edge]



from optinist.api.edit_ROI.wrappers.caiman_edit_roi.add_roi import execute_add_ROI
from optinist.api.edit_ROI.wrappers.caiman_edit_roi.delete_roi import excute_delete_roi
from optinist.api.edit_ROI.wrappers.caiman_edit_roi.merge_roi import execute_merge_roi

__all__ = ["execute_add_ROI", "excute_delete_roi", "execute_merge_roi"]



from optinist.wrappers.dummy_wrapper.dummy import (
    dummy_image2heat,
    dummy_image2image,
    dummy_image2image8roi,
    dummy_image2image8roi8time8heat,
    dummy_image2image8time,
    dummy_image2roi,
    dummy_image2scatter,
    dummy_image2time,
    dummy_image2time8iscell,
    dummy_keyerror,
    dummy_time2time,
    dummy_typeerror,
)

dummy_wrapper_dict = {
    "dummy": {
        "dummy_image2image": {
            "function": dummy_image2image,
        },
        "dummy_image2time": {
            "function": dummy_image2time,
        },
        "dummy_image2heat": {
            "function": dummy_image2heat,
        },
        "dummy_time2time": {
            "function": dummy_time2time,
        },
        "dummy_image2image8time": {
            "function": dummy_image2image8time,
        },
        "dummy_keyerror": {
            "function": dummy_keyerror,
        },
        "dummy_typeerror": {
            "function": dummy_typeerror,
        },
        "dummy_image2time8iscell": {
            "function": dummy_image2time8iscell,
        },
        "dummy_image2roi": {
            "function": dummy_image2roi,
        },
        "dummy_image2image8roi": {
            "function": dummy_image2image8roi,
        },
        "dummy_image2image8roi8time8heat": {
            "function": dummy_image2image8roi8time8heat,
        },
        "dummy_image2scatter": {
            "function": dummy_image2scatter,
        },
    }
}



from dataclasses import asdict
from typing import Dict, List

from optinist.api.experiment.experiment_reader import ExptConfigReader
from optinist.api.experiment.experiment_writer import ExptConfigWriter
from optinist.api.snakemake.smk import FlowConfig, Rule, SmkParam
from optinist.api.snakemake.snakemake_executor import (
    delete_dependencies,
    snakemake_execute,
)
from optinist.api.snakemake.snakemake_reader import SmkParamReader
from optinist.api.snakemake.snakemake_rule import SmkRule
from optinist.api.snakemake.snakemake_writer import SmkConfigWriter
from optinist.api.workflow.workflow import NodeType, RunItem
from optinist.api.workflow.workflow_params import get_typecheck_params


class WorkflowRunner:
    def __init__(self, unique_id: str, runItem: RunItem) -> None:
        self.unique_id = unique_id
        self.runItem = runItem
        self.nodeDict = ExptConfigReader.read_nodeDict(self.runItem.nodeDict)
        self.edgeDict = ExptConfigReader.read_edgeDict(self.runItem.edgeDict)

        ExptConfigWriter(
            self.unique_id, self.runItem.name, self.nodeDict, self.edgeDict
        ).write()

    def run_workflow(self, background_tasks):
        self.set_smk_config()

        snakemake_params: SmkParam = get_typecheck_params(
            self.runItem.snakemakeParam, "snakemake"
        )
        snakemake_params = SmkParamReader.read(snakemake_params)
        snakemake_params.forcerun = self.runItem.forceRunList
        if len(snakemake_params.forcerun) > 0:
            delete_dependencies(
                unique_id=self.unique_id,
                smk_params=snakemake_params,
                nodeDict=self.nodeDict,
                edgeDict=self.edgeDict,
            )
        background_tasks.add_task(snakemake_execute, self.unique_id, snakemake_params)

    def set_smk_config(self):
        rules, last_output = self.rulefile()

        flow_config = FlowConfig(
            rules=rules,
            last_output=last_output,
        )

        SmkConfigWriter.write(self.unique_id, asdict(flow_config))

    def rulefile(self):
        endNodeList = self.get_endNodeList()

        nwbfile = get_typecheck_params(self.runItem.nwbParam, "nwb")

        rule_dict: Dict[str, Rule] = {}
        last_outputs = []

        for node in self.nodeDict.values():
            if node.type == NodeType.IMAGE:
                rule_dict[node.id] = SmkRule(
                    unique_id=self.unique_id,
                    node=node,
                    edgeDict=self.edgeDict,
                    nwbfile=nwbfile,
                ).image()
            elif node.type == NodeType.CSV:
                rule_dict[node.id] = SmkRule(
                    unique_id=self.unique_id,
                    node=node,
                    edgeDict=self.edgeDict,
                    nwbfile=nwbfile,
                ).csv()
            elif node.type == NodeType.FLUO:
                rule_dict[node.id] = SmkRule(
                    unique_id=self.unique_id,
                    node=node,
                    edgeDict=self.edgeDict,
                    nwbfile=nwbfile,
                ).csv()
            elif node.type == NodeType.BEHAVIOR:
                rule_dict[node.id] = SmkRule(
                    unique_id=self.unique_id,
                    node=node,
                    edgeDict=self.edgeDict,
                    nwbfile=nwbfile,
                ).csv(nodeType="behavior")
            elif node.type == NodeType.HDF5:
                rule_dict[node.id] = SmkRule(
                    unique_id=self.unique_id,
                    node=node,
                    edgeDict=self.edgeDict,
                    nwbfile=nwbfile,
                ).hdf5()
            elif node.type == NodeType.ALGO:
                rule = SmkRule(
                    unique_id=self.unique_id,
                    node=node,
                    edgeDict=self.edgeDict,
                ).algo(nodeDict=self.nodeDict)

                rule_dict[node.id] = rule

                if node.id in endNodeList:
                    last_outputs.append(rule.output)
            else:
                assert False, "NodeType doesn't exists"

        return rule_dict, last_outputs

    def get_endNodeList(self) -> List[str]:
        returnCntDict = {key: 0 for key in self.nodeDict.keys()}
        for edge in self.edgeDict.values():
            returnCntDict[edge.source] += 1

        endNodeList = []
        for key, value in returnCntDict.items():
            if value == 0:
                endNodeList.append(key)
        return endNodeList



from optinist.api.dataclass.dataclass import FluoData, IscellData, ScatterData
from optinist.api.nwb.nwb import NWBDATASET
from optinist.wrappers.optinist_wrapper.utils import standard_norm


def TSNE(
    neural_data: FluoData,
    output_dir: str,
    iscell: IscellData = None,
    params: dict = None,
) -> dict():
    import numpy as np
    from sklearn.manifold import TSNE

    neural_data = neural_data.data

    # data shold be time x component matrix
    if params["transpose"]:
        X = neural_data.transpose()
    else:
        X = neural_data

    if iscell is not None:
        iscell = iscell.data
        ind = np.where(iscell > 0)[0]
        X = X[:, ind]

    # preprocessing
    tX = standard_norm(X, params["standard_mean"], params["standard_std"])

    # calculate TSNE
    tsne = TSNE(**params["TSNE"])

    proj_X = tsne.fit_transform(tX)

    # NWB追加
    nwbfile = {}
    nwbfile[NWBDATASET.POSTPROCESS] = {"projectedNd": proj_X}

    info = {
        "projectedNd": ScatterData(proj_X, file_name="projectedNd"),
        "nwbfile": nwbfile,
    }

    return info



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass as pydantic_dataclass


@dataclass
class Arg:
    name: str
    type: str
    isNone: bool


@dataclass
class Return:
    name: str
    type: str


@dataclass
class Algo:
    args: List[Arg]
    returns: List[Return]
    parameter: str = None
    path: str = None


@pydantic_dataclass
class TreeNode:
    path: str
    name: str
    isdir: bool
    nodes: List["TreeNode"]


@dataclass
class FILETYPE:
    IMAGE: str = "image"
    CSV: str = "csv"
    HDF5: str = "hdf5"
    BEHAVIOR: str = "behavior"


class DeleteItem(BaseModel):
    uidList: list


class RoiPos(BaseModel):
    posx: int
    posy: int
    sizex: int
    sizey: int


class RoiList(BaseModel):
    ids: List[int] = Field(default=[0, 1])


class EditRoiSuccess(BaseModel):
    max_index: int


@pydantic_dataclass
class HDF5Node:
    isDir: bool
    name: str
    path: str
    nodes: List["HDF5Node"] = None
    shape: tuple = None
    nbytes: str = None


@dataclass
class OutputData:
    data: Union[List, Dict, str]
    columns: List[str] = None
    index: List[str] = None


@dataclass
class JsonTimeSeriesData(OutputData):
    xrange: list = None
    std: Dict[str, dict] = None


@dataclass
class FilePath:
    file_path: str


class AlgoModel(BaseModel):
    children: Union[Dict[str, Algo], Dict[str, "AlgoModel"]]


class AlgoList(BaseModel):
    __root__: Dict[str, AlgoModel] = {
        "caiman": {
            "children": {
                "caiman_mc": {
                    "args": [{"name": "image", "type": "ImageData", "isNone": False}],
                    "returns": [{"name": "mc_images", "type": "ImageData"}],
                    "parameter": None,
                    "path": "caiman/caiman_mc",
                }
            }
        }
    }


class NWBParams(BaseModel):
    session_description: str = "optinist"
    identifier: str = "optinist"
    experiment_description: Optional[str] = None
    device: Union[Dict, Any]
    optical_channel: Union[Dict, Any]
    imaging_plane: Union[Dict, Any]
    image_series: Union[Dict, Any]
    ophys: Union[Dict, Any]


class SnakemakeParams(BaseModel):
    use_conda: bool
    cores: int
    forceall: bool
    forcetargets: bool
    lock: bool


class RenameItem(BaseModel):
    new_name: str



import gc

import numpy as np

from optinist.api.dataclass.dataclass import (
    CaimanCnmfData,
    FluoData,
    ImageData,
    IscellData,
    RoiData,
)
from optinist.api.nwb.nwb import NWBDATASET
from optinist.api.utils.filepath_creater import join_filepath


def get_roi(A, thr, thr_method, swap_dim, dims):
    from scipy.ndimage import binary_fill_holes
    from skimage.measure import find_contours

    d, nr = np.shape(A)

    # for each patches
    ims = []
    coordinates = []
    for i in range(nr):
        pars = dict()
        # we compute the cumulative sum of the energy of the Ath component
        # that has been ordered from least to highest
        patch_data = A.data[A.indptr[i] : A.indptr[i + 1]]
        indx = np.argsort(patch_data)[::-1]

        if thr_method == "nrg":
            cumEn = np.cumsum(patch_data[indx] ** 2)
            if len(cumEn) == 0:
                pars = dict(
                    coordinates=np.array([]),
                    CoM=np.array([np.NaN, np.NaN]),
                    neuron_id=i + 1,
                )
                coordinates.append(pars)
                continue
            else:
                # we work with normalized values
                cumEn /= cumEn[-1]
                Bvec = np.ones(d)
                # we put it in a similar matrix
                Bvec[A.indices[A.indptr[i] : A.indptr[i + 1]][indx]] = cumEn
        else:
            Bvec = np.zeros(d)
            Bvec[A.indices[A.indptr[i] : A.indptr[i + 1]]] = (
                patch_data / patch_data.max()
            )

        if swap_dim:
            Bmat = np.reshape(Bvec, dims, order="C")
        else:
            Bmat = np.reshape(Bvec, dims, order="F")

        r_mask = np.zeros_like(Bmat, dtype="bool")
        contour = find_contours(Bmat, thr)
        for c in contour:
            r_mask[np.round(c[:, 0]).astype("int"), np.round(c[:, 1]).astype("int")] = 1

        # Fill in the hole created by the contour boundary
        r_mask = binary_fill_holes(r_mask)
        ims.append(r_mask + (i * r_mask))

    return ims


def caiman_cnmf(
    images: ImageData, output_dir: str, params: dict = None
) -> dict(fluorescence=FluoData, iscell=IscellData):
    import scipy
    from caiman import local_correlations, stop_server
    from caiman.cluster import setup_cluster
    from caiman.mmapping import prepare_shape
    from caiman.paths import memmap_frames_filename
    from caiman.source_extraction.cnmf import cnmf
    from caiman.source_extraction.cnmf.params import CNMFParams

    file_path = images.path
    if isinstance(file_path, list):
        file_path = file_path[0]

    images = images.data

    # np.arrayをmmapへ変換
    order = "C"
    dims = images.shape[1:]
    T = images.shape[0]
    shape_mov = (np.prod(dims), T)

    dir_path = join_filepath(file_path.split("/")[:-1])
    basename = file_path.split("/")[-1]
    fname_tot = memmap_frames_filename(basename, dims, T, order)

    mmap_images = np.memmap(
        join_filepath([dir_path, fname_tot]),
        mode="w+",
        dtype=np.float32,
        shape=prepare_shape(shape_mov),
        order=order,
    )

    mmap_images = np.reshape(mmap_images.T, [T] + list(dims), order="F")
    mmap_images[:] = images[:]

    del images
    gc.collect()

    if params is None:
        ops = CNMFParams()
    else:
        ops = CNMFParams(params_dict=params)

    if "dview" in locals():
        stop_server(dview=dview)  # noqa: F821

    c, dview, n_processes = setup_cluster(
        backend="local", n_processes=None, single_thread=True
    )

    cnm = cnmf.CNMF(n_processes=n_processes, dview=dview, Ain=None, params=ops)
    cnm = cnm.fit(mmap_images)

    stop_server(dview=dview)

    # contours plot
    Cn = local_correlations(mmap_images.transpose(1, 2, 0))
    Cn[np.isnan(Cn)] = 0

    thr = params["thr"]
    thr_method = "nrg"
    swap_dim = False

    iscell = np.concatenate(
        [np.ones(cnm.estimates.A.shape[-1]), np.zeros(cnm.estimates.b.shape[-1])]
    )

    ims = get_roi(cnm.estimates.A, thr, thr_method, swap_dim, dims)
    ims = np.stack(ims)
    cell_roi = np.nanmax(ims, axis=0).astype(float)
    cell_roi[cell_roi == 0] = np.nan

    non_cell_roi_ims = get_roi(
        scipy.sparse.csc_matrix(cnm.estimates.b), thr, thr_method, swap_dim, dims
    )
    non_cell_roi_ims = np.stack(non_cell_roi_ims)
    non_cell_roi = np.nanmax(non_cell_roi_ims, axis=0).astype(float)
    non_cell_roi[non_cell_roi == 0] = np.nan

    all_roi = np.nanmax(np.stack([cell_roi, non_cell_roi]), axis=0)

    # NWBの追加
    nwbfile = {}
    # NWBにROIを追加
    roi_list = []
    n_cells = cnm.estimates.A.shape[-1]
    for i in range(n_cells):
        kargs = {}
        kargs["image_mask"] = cnm.estimates.A.T[i].T.toarray().reshape(dims)
        if hasattr(cnm.estimates, "accepted_list"):
            kargs["accepted"] = i in cnm.estimates.accepted_list
        if hasattr(cnm.estimates, "rejected_list"):
            kargs["rejected"] = i in cnm.estimates.rejected_list
        roi_list.append(kargs)

    # backgroundsを追加
    bg_list = []
    for bg in cnm.estimates.b.T:
        kargs = {}
        kargs["image_mask"] = bg.reshape(dims)
        if hasattr(cnm.estimates, "accepted_list"):
            kargs["accepted"] = False
        if hasattr(cnm.estimates, "rejected_list"):
            kargs["rejected"] = False
        bg_list.append(kargs)

    nwbfile[NWBDATASET.ROI] = {
        "roi_list": roi_list,
        "bg_list": bg_list,
    }

    # iscellを追加
    nwbfile[NWBDATASET.COLUMN] = {
        "roi_column": {
            "name": "iscell",
            "discription": "two columns - iscell & probcell",
            "data": iscell,
        }
    }

    # Fluorescence
    n_rois = cnm.estimates.A.shape[-1]
    n_bg = len(cnm.estimates.f)

    nwbfile[NWBDATASET.FLUORESCENCE] = {
        "RoiResponseSeries": {
            "table_name": "ROIs",
            "region": list(range(n_rois)),
            "name": "RoiResponseSeries",
            "data": cnm.estimates.C.T,
            "unit": "lumens",
        },
        "Background_Fluorescence_Response": {
            "table_name": "Background",
            "region": list(range(n_rois, n_rois + n_bg)),
            "name": "Background_Fluorescence_Response",
            "data": cnm.estimates.f.T,
            "unit": "lumens",
        },
    }

    fluorescence = np.concatenate(
        [
            cnm.estimates.C,
            cnm.estimates.f,
        ]
    )

    cnmf_data = {}
    cnmf_data["fluorescence"] = fluorescence
    cnmf_data["im"] = np.concatenate([ims, non_cell_roi_ims], axis=0)
    cnmf_data["is_cell"] = iscell.astype(bool)
    cnmf_data["images"] = mmap_images

    info = {
        "images": ImageData(
            np.array(Cn * 255, dtype=np.uint8),
            output_dir=output_dir,
            file_name="images",
        ),
        "fluorescence": FluoData(fluorescence, file_name="fluorescence"),
        "iscell": IscellData(iscell, file_name="iscell"),
        "all_roi": RoiData(all_roi, output_dir=output_dir, file_name="all_roi"),
        "cell_roi": RoiData(cell_roi, output_dir=output_dir, file_name="cell_roi"),
        "non_cell_roi": RoiData(
            non_cell_roi, output_dir=output_dir, file_name="non_cell_roi"
        ),
        "nwbfile": nwbfile,
        "cnmf_data": CaimanCnmfData(cnmf_data),
    }

    return info



from optinist.api.dataclass.dataclass import CaimanCnmfData, RoiData
from optinist.api.edit_ROI.wrappers.caiman_edit_roi.utils import set_nwbfile


def excute_delete_roi(node_dirpath, ids):
    import numpy as np

    # load data
    cnmf_data = np.load(f"{node_dirpath}/caiman_cnmf.npy", allow_pickle=True).item()
    is_cell = cnmf_data.get("is_cell")
    delete_roi = cnmf_data.get("delete_roi", [])
    im = cnmf_data.get("im")

    # delete ROI
    is_cell[ids] = False
    [delete_roi.append(id + 1) for id in ids]

    cell_roi = np.zeros(im.shape)
    num_rois = im.shape[0]
    for i in range(num_rois):
        cell_roi[i, :, :] = np.where(im[i, :, :] != 0, i + 1, np.nan)

    cnmf_data["is_cell"] = is_cell
    cnmf_data["delete_roi"] = delete_roi
    info = {
        "cell_roi": RoiData(
            np.nanmax(cell_roi[is_cell], axis=0),
            output_dir=node_dirpath,
            file_name="cell_roi",
        ),
        "cnmf_data": CaimanCnmfData(cnmf_data),
        "nwbfile": set_nwbfile(cnmf_data),
    }
    return info



from optinist.api.dataclass.dataclass import (
    FluoData,
    ImageData,
    IscellData,
    RoiData,
    Suite2pData,
)
from optinist.api.nwb.nwb import NWBDATASET


def suite2p_roi(
    ops: Suite2pData, output_dir: str, params: dict = None
) -> dict(ops=Suite2pData, fluorescence=FluoData, iscell=IscellData):
    import numpy as np
    from suite2p import ROI, classification, default_ops, detection, extraction

    print("start suite2p_roi")
    ops = ops.data

    ops = {**default_ops(), **ops, **params}

    # ROI detection
    ops_classfile = ops.get("classifier_path")
    builtin_classfile = classification.builtin_classfile
    user_classfile = classification.user_classfile
    if ops_classfile:
        print(f"NOTE: applying classifier {str(ops_classfile)}")
        classfile = ops_classfile
    elif ops["use_builtin_classifier"] or not user_classfile.is_file():
        print(f"NOTE: Applying builtin classifier at {str(builtin_classfile)}")
        classfile = builtin_classfile
    else:
        print(f"NOTE: applying default {str(user_classfile)}")
        classfile = user_classfile

    ops, stat = detection.detect(ops=ops, classfile=classfile)

    # ROI EXTRACTION
    ops, stat, F, Fneu, _, _ = extraction.create_masks_and_extract(ops, stat)
    stat = stat.tolist()

    # ROI CLASSIFICATION
    iscell = classification.classify(stat=stat, classfile=classfile)
    iscell = iscell[:, 0].astype(bool)

    arrays = []
    for i, s in enumerate(stat):
        array = ROI(
            ypix=s["ypix"], xpix=s["xpix"], lam=s["lam"], med=s["med"], do_crop=False
        ).to_array(Ly=ops["Ly"], Lx=ops["Lx"])
        array *= i + 1
        arrays.append(array)

    im = np.stack(arrays)
    im[im == 0] = np.nan

    # roiを追加
    roi_list = []
    for i in range(len(stat)):
        kargs = {}
        kargs["pixel_mask"] = np.array(
            [stat[i]["ypix"], stat[i]["xpix"], stat[i]["lam"]]
        ).T
        roi_list.append(kargs)

    # NWBを追加
    nwbfile = {}

    nwbfile[NWBDATASET.ROI] = {"roi_list": roi_list}

    # iscellを追加
    nwbfile[NWBDATASET.COLUMN] = {
        "roi_column": {
            "name": "iscell",
            "discription": "two columns - iscell & probcell",
            "data": iscell,
        }
    }

    # Fluorenceを追加
    nwbfile[NWBDATASET.FLUORESCENCE] = {}
    for name, data in zip(["Fluorescence", "Neuropil"], [F, Fneu]):
        nwbfile[NWBDATASET.FLUORESCENCE][name] = {
            "table_name": name,
            "region": list(range(len(data))),
            "name": name,
            "data": data,
            "unit": "lumens",
            "rate": ops["fs"],
        }

    ops["stat"] = stat
    ops["F"] = F
    ops["Fneu"] = Fneu
    ops["iscell"] = iscell
    ops["im"] = im

    info = {
        "ops": Suite2pData(ops),
        "max_proj": ImageData(
            ops["max_proj"], output_dir=output_dir, file_name="max_proj"
        ),
        "Vcorr": ImageData(ops["Vcorr"], output_dir=output_dir, file_name="Vcorr"),
        "fluorescence": FluoData(F, file_name="fluorescence"),
        "iscell": IscellData(iscell, file_name="iscell"),
        "all_roi": RoiData(
            np.nanmax(im, axis=0), output_dir=output_dir, file_name="all_roi"
        ),
        "non_cell_roi": RoiData(
            np.nanmax(im[~iscell], axis=0),
            output_dir=output_dir,
            file_name="noncell_roi",
        ),
        "cell_roi": RoiData(
            np.nanmax(im[iscell], axis=0), output_dir=output_dir, file_name="cell_roi"
        ),
        "nwbfile": nwbfile,
    }

    return info



import inspect
from typing import Dict, List, ValuesView

from fastapi import APIRouter

from optinist.routers.const import NOT_DISPLAY_ARGS_LIST
from optinist.routers.model import Algo, AlgoList, Arg, Return
from optinist.wrappers import wrapper_dict

router = APIRouter()


class NestDictGetter:
    @classmethod
    def get_nest_dict(cls, parent_value, parent_key: str) -> Dict[str, Algo]:
        algo_dict = {}
        for key, value in parent_value.items():
            algo_dict[key] = {}
            if isinstance(value, dict) and "function" not in value:
                algo_dict[key]["children"] = cls.get_nest_dict(
                    value, cls._parent_key(parent_key, key)
                )
            else:
                sig = inspect.signature(value["function"])
                returns_list = None
                if sig.return_annotation is not inspect._empty:
                    returns_list = cls._return_list(sig.return_annotation.items())

                algo_dict[key] = Algo(
                    args=cls._args_list(sig.parameters.values()),
                    returns=returns_list,
                    parameter=value["parameter"] if "parameter" in value else None,
                    path=cls._parent_key(parent_key, key),
                )

        return algo_dict

    @classmethod
    def _args_list(cls, arg_params: ValuesView[inspect.Parameter]) -> List[Arg]:
        return [
            Arg(
                name=x.name,
                type=x.annotation.__name__,
                isNone=x.default is None,
            )
            for x in arg_params
            if x.name not in NOT_DISPLAY_ARGS_LIST
        ]

    @classmethod
    def _return_list(cls, return_params: ValuesView[inspect.Parameter]) -> List[Return]:
        return [Return(name=k, type=v.__name__) for k, v in return_params]

    @classmethod
    def _parent_key(cls, parent_key: str, key: str) -> str:
        if parent_key == "":
            return key
        else:
            return f"{parent_key}/{key}"


@router.get("/algolist", response_model=AlgoList, tags=["others"])
async def get_algolist() -> Dict[str, Algo]:
    """_summary_

    Returns:
        {
            'caiman': {
                'children': {
                    'caiman_mc' : {
                        'args': ['images', 'timeseries'],
                        'return': ['images'],
                        'path': 'caiman/caiman_mc'
                    },
                    'caiman_cnmf': {
                        'args': ['images', 'timeseries'],
                        'return': ['images'],
                        'path': 'caiman/caiman_mc'
                    }
                }
            }
        }
    """

    return NestDictGetter.get_nest_dict(wrapper_dict, "")



import os

import numpy as np

from optinist.api.dataclass.dataclass import FluoData, RoiData, Suite2pData
from optinist.api.edit_ROI.wrappers.suite2p_edit_roi.utils import set_nwbfile


def execute_merge_roi(node_dirpath, ids):
    from suite2p import ROI
    from suite2p.detection.stats import median_pix, roi_stats

    ops = np.load(os.path.join(node_dirpath, "suite2p.npy"), allow_pickle=True).item()
    iscell = ops.get("iscell")
    stat_orig = ops.get("stat")
    im = ops.get("im")
    F_all = ops["F"]
    Fneu_all = ops["Fneu"]
    merge_roi = ops.get("merge_roi", [])

    print("merging activity... this may take some time")
    ypix = np.zeros((0,), np.int32)
    xpix = np.zeros((0,), np.int32)
    lam = np.zeros((0,), np.float32)
    footprints = np.array([])
    F = np.zeros((0, F_all.shape[1]), np.float32)
    Fneu = np.zeros((0, F_all.shape[1]), np.float32)

    # merged cells id
    merged_cells = []
    for n in np.array(ids):
        merged_cells.append(n)
    merged_cells = np.unique(np.array(merged_cells))

    # merge cells regions
    for n in merged_cells:
        # deactivate merged cells
        iscell[n] = False
        ypix = np.append(ypix, stat_orig[n]["ypix"])
        xpix = np.append(xpix, stat_orig[n]["xpix"])
        lam = np.append(lam, stat_orig[n]["lam"])
        footprints = np.append(footprints, stat_orig[n]["footprint"])
        F = np.append(F, F_all[n, :][np.newaxis, :], axis=0)
        Fneu = np.append(Fneu, Fneu_all[n, :][np.newaxis, :], axis=0)

    # remove overlaps from merged cells regions
    ipix = np.concatenate((ypix[:, np.newaxis], xpix[:, np.newaxis]), axis=1)
    _, goodi = np.unique(ipix, return_index=True, axis=0)
    ypix = ypix[goodi]
    xpix = xpix[goodi]
    lam = lam[goodi]

    # compute statistics of merges
    stat0 = {}
    stat0["ypix"] = ypix
    stat0["xpix"] = xpix
    stat0["lam"] = lam / lam.sum()
    stat0["npix"] = ypix.size
    stat0["med"] = median_pix(ypix, xpix)

    stat0["chan2_prob"] = -1
    stat0["inmerge"] = -1

    # compute activity of merged cells
    F = F.mean(axis=0)
    Fneu = Fneu.mean(axis=0)

    if "aspect" in ops:
        d0 = np.array([int(ops["aspect"] * 10), 10])
    else:
        d0 = ops["diameter"]
        if isinstance(d0, int):
            d0 = [d0, d0]

    # add cell to structs
    print(type(stat_orig), type(stat0), type(stat_orig[0]))
    stat_orig.append(stat0)
    stat_orig = roi_stats(stat_orig, d0[0], d0[1], ops["Ly"], ops["Lx"])
    stat_orig[-1]["lam"] = stat_orig[-1]["lam"] * merged_cells.size

    array = np.expand_dims(
        ROI(
            ypix=stat_orig[-1]["ypix"],
            xpix=stat_orig[-1]["xpix"],
            lam=stat_orig[-1]["lam"],
            med=stat_orig[-1]["med"],
            do_crop=False,
        ).to_array(Ly=ops["Ly"], Lx=ops["Lx"]),
        axis=0,
    )
    id = np.nanmax(im) + 1
    array *= id
    merge_roi.append(float(id))
    merge_roi += (merged_cells + 1).tolist()
    merge_roi.append((-1.0))

    im = np.concatenate((im, array), axis=0)
    im[im == 0] = np.nan

    try:
        F_all = np.concatenate((ops["F"], np.expand_dims(F, axis=0)), axis=0)
        Fneu_all = np.concatenate((ops["Fneu"], np.expand_dims(Fneu, axis=0)), axis=0)
        iscell = np.append(iscell, True)
    except Exception as e:
        print(e)

    ops["stat"] = stat_orig
    ops["F"] = F_all
    ops["Fneu"] = Fneu_all
    ops["iscell"] = iscell
    ops["im"] = im
    ops["merge_roi"] = merge_roi

    info = {
        "ops": Suite2pData(ops),
        "fluorescence": FluoData(ops["F"], file_name="fluorescence"),
        "all_roi": RoiData(
            np.nanmax(im, axis=0), output_dir=node_dirpath, file_name="all_roi"
        ),
        "non_cell_roi": RoiData(
            np.nanmax(im[~iscell], axis=0),
            output_dir=node_dirpath,
            file_name="noncell_roi",
        ),
        "cell_roi": RoiData(
            np.nanmax(im[iscell], axis=0), output_dir=node_dirpath, file_name="cell_roi"
        ),
        "nwbfile": set_nwbfile(ops),
    }
    return info



import os
from collections import deque
from typing import Dict

from snakemake import snakemake

from optinist.api.dir_path import DIRPATH
from optinist.api.logger import Logger
from optinist.api.snakemake.smk import SmkParam
from optinist.api.utils.filepath_creater import get_pickle_file, join_filepath
from optinist.api.workflow.workflow import Edge, Node


def snakemake_execute(unique_id: str, params: SmkParam):
    snakemake(
        DIRPATH.SNAKEMAKE_FILEPATH,
        forceall=params.forceall,
        cores=params.cores,
        use_conda=params.use_conda,
        workdir=f"{os.path.dirname(DIRPATH.ROOT_DIR)}",
        log_handler=[Logger(unique_id).smk_logger],
    )


def delete_dependencies(
    unique_id: str,
    smk_params: SmkParam,
    nodeDict: Dict[str, Node],
    edgeDict: Dict[str, Edge],
):
    queue = deque()

    for param in smk_params.forcerun:
        queue.append(param.nodeId)

    while True:
        # terminate condition
        if len(queue) == 0:
            break

        # delete pickle
        node_id = queue.pop()
        algo_name = nodeDict[node_id].data.label

        pickle_filepath = join_filepath(
            [
                DIRPATH.OUTPUT_DIR,
                get_pickle_file(
                    unique_id=unique_id,
                    node_id=node_id,
                    algo_name=algo_name,
                ),
            ]
        )
        # print(pickle_filepath)

        if os.path.exists(pickle_filepath):
            os.remove(pickle_filepath)

        # 全てのedgeを見て、node_idがsourceならtargetをqueueに追加する
        for edge in edgeDict.values():
            if node_id == edge.source:
                queue.append(edge.target)



from optinist.api.nwb.nwb import NWBDATASET


def set_nwbfile(lccd_data, roi_list, fluorescence=None):
    nwbfile = {}

    nwbfile[NWBDATASET.ROI] = {"roi_list": roi_list}

    if fluorescence is not None:
        nwbfile[NWBDATASET.FLUORESCENCE] = {}
        nwbfile[NWBDATASET.FLUORESCENCE]["Fluorescence"] = {
            "table_name": "Fluorescence",
            "region": list(range(len(fluorescence))),
            "name": "Fluorescence",
            "data": fluorescence,
            "unit": "lumens",
        }

    nwbfile[NWBDATASET.COLUMN] = {
        "roi_column": {
            "name": "iscell",
            "discription": "two columns - iscell & probcell",
            "data": lccd_data.get("is_cell"),
        }
    }

    # NWB追加
    nwbfile[NWBDATASET.POSTPROCESS] = {
        "add_roi": lccd_data.get("add_roi", []),
        "delete_roi": lccd_data.get("delete_roi", []),
        "merge_roi": lccd_data.get("merge_roi", []),
    }

    return nwbfile
